import { Query } from ".";
import { Archetype } from "./Archetype";
/**
 * A base interface for entities, which are just normal JavaScript objects with
 * any number of properties, each of which represents a single component. When
 * using hmecs, it is recommended to create your own type representing your
 * game's entities, and pass that to `createWorld` for full typing support.
 */
export interface IEntity {
    id?: number;
}
/**
 * For situations where no entity type argument is passed to createWorld, we'll
 * default to an untyped entity type that can hold any component.
 */
export declare type UntypedEntity = {
    [components: string]: ComponentData;
} & IEntity;
/**
 * Component names are just strings/object property names.
 */
export declare type ComponentName<T extends IEntity> = keyof T;
/**
 * The data of a component can be literally anything. Good times!
 */
export declare type ComponentData = any;
/**
 * A tag is just an "empty" component. For convenience and nicer type support, we're
 * providing a Tag type and constant, both of which are simply equal to `true`.
 */
export declare const Tag = true;
export declare type Tag = true;
export declare class World<T extends IEntity = UntypedEntity> {
    /** An array holding all entities known to this world. */
    entities: T[];
    /** An ID generator we use for assigning IDs to newly added entities. */
    private nextId;
    /** A list of known archetypes. */
    private archetypes;
    archetype<TQuery extends Query<T>>(...query: TQuery): Archetype<T, TQuery>;
    private indexEntity;
    createEntity: (entity?: T) => T;
    destroyEntity: (entity: T) => void;
    addComponent: <U extends keyof T>(entity: T, name: U, data: T[U]) => void;
    removeComponent: (entity: T, ...components: ComponentName<T>[]) => void;
    /**
     * A queue of commands to run.
     */
    private queuedCommands;
    queue: {
        createEntity: (entity: T) => T;
        destroyEntity: (entity: T) => void;
        addComponent: <U extends keyof T>(entity: T, name: U, data: T[U]) => void;
        removeComponent: (entity: T, ...names: ComponentName<T>[]) => void;
        flush: () => void;
    };
    /**
     * Clear the entire world by discarding all entities and indices.
     */
    clear(): void;
}
