import { jsx as _jsx, Fragment as _Fragment } from "react/jsx-runtime";
import { createContext, useContext, useEffect, memo, cloneElement, useRef } from "react";
import { World, Tag } from ".";
import { useData } from "./util/useData";
import { useRerender } from "./util/useRerender";
export function createECS() {
    const world = new World();
    const EntityContext = createContext(null);
    /**
     * A React component to either create a new entity, or represent an existing entity so
     * it can be enhanced with additional components (see the <Component> component.)
     */
    const Entity = ({ entity: existingEntity, children }) => {
        /* Reuse the specified entity or create a new one */
        const entity = useData(() => existingEntity !== null && existingEntity !== void 0 ? existingEntity : world.createEntity());
        /* If the entity was freshly created, manage its presence in the ECS world. */
        useEffect(() => {
            if (existingEntity)
                return;
            return () => world.destroyEntity(entity);
        }, [entity]);
        /* Provide a context with the entity so <Component> components can be wired up. */
        return (_jsx(EntityContext.Provider, Object.assign({ value: entity }, { children: typeof children === "function" ? children(entity) : children })));
    };
    const MemoizedEntity = memo(({ entity, children }) => (_jsx(Entity, Object.assign({ entity: entity }, { children: typeof children === "function" ? children(entity) : children }), entity.id)), (a, b) => a.entity === b.entity);
    const Entities = ({ entities, memoize = false, children }) => {
        const Klass = memoize ? MemoizedEntity : Entity;
        return (_jsx(_Fragment, { children: entities.map((entity) => (_jsx(Klass, { entity: entity, children: children }, entity.id))) }));
    };
    const Collection = ({ initial = 0, memoize = false, tag, children }) => {
        const { entities } = useArchetype(tag);
        useEffect(() => {
            const ents = [];
            /* When firing up, create the requested number of entities. */
            for (let i = 0; i < initial; i++) {
                const entity = world.createEntity();
                world.addComponent(entity, tag, Tag); /* oh no */
                ents.push(entity);
            }
            /* When shutting down, purge all the entities we have created. */
            return () => {
                for (const entity of entities) {
                    world.destroyEntity(entity);
                }
            };
        }, [tag, initial]);
        return (_jsx(Entities, Object.assign({ entities: entities, memoize: memoize }, { children: children })));
    };
    /**
     * Return the current entity context.
     */
    function useEntity() {
        return useContext(EntityContext);
    }
    /**
     * Declaratively add a component to an entity.
     */
    function Component(props) {
        const entity = useEntity();
        const ref = useRef();
        /* Warn the user that passing multiple children is not allowed. */
        if ("children" in props && Array.isArray(props.children)) {
            throw new Error("<Component> will only accept a single React child.");
        }
        useEffect(() => {
            world.addComponent(entity, props.name, "data" in props ? props.data : ref.current);
            return () => {
                /* The entity might already have been destroyed, so let's check. */
                if ("id" in entity) {
                    world.removeComponent(entity, props.name);
                }
            };
        }, [entity, props.name, "data" in props && props.data]);
        return (_jsx(_Fragment, { children: "children" in props &&
                cloneElement(typeof props.children === "function" ? props.children(entity) : props.children, { ref }) }));
    }
    /**
     * Return the entities of the specified archetype and subscribe this component
     * to it, making it re-render when entities are added to or removed from it.
     */
    function useArchetype(...query) {
        const rerender = useRerender();
        const archetype = useData(() => world.archetype(...query));
        useEffect(() => {
            archetype.onEntityAdded.on(rerender);
            archetype.onEntityRemoved.on(rerender);
            /* We need to rerender at least once, because other effects might have set up
               new entities before we had a chance to register our listeners. */
            rerender();
            return () => {
                archetype.onEntityAdded.off(rerender);
                archetype.onEntityRemoved.off(rerender);
            };
        }, [archetype]);
        return archetype;
    }
    return {
        world,
        useArchetype,
        useEntity,
        Entity,
        Component,
        MemoizedEntity,
        Entities,
        Collection
    };
}
